import json
import os
import shutil
import warnings

from gi.repository import Gio
from gi.repository import GLib

from gsem.utils import get_json_response
from gsem.utils import download_and_extract_zip
from gsem.config import (
    EXTENSION_DIR,
    GNOME_SHELL_VERSION,
    API_ROOT,
    API_DETAIL,
    API_SEARCH,
)


class Extension:

    def __init__(self, uuid):
        self.uuid = uuid
        self._meta = None
        self._remote_meta = None

    @property
    def meta(self):
        """Local metadata dict."""
        if not self._meta:
            with open(os.path.join(EXTENSION_DIR, self.uuid, 'metadata.json')) as f:
                self._meta = json.load(f)
        return self._meta

    @property
    def remote_meta(self):
        """Remote metadata dict."""
        if not self._remote_meta:
            self._remote_meta = get_json_response(API_DETAIL, {
                'uuid': self.uuid,
                'shell_version': '.'.join(str(v) for v in GNOME_SHELL_VERSION),
            })
        return self._remote_meta

    @remote_meta.setter
    def remote_meta(self, value):
        self.uuid = value['uuid']
        self._remote_meta = value

    def outdated(self):
        # TODO: use https://extensions.gnome.org/update-info/?installed={%22arch-update@RaphaelRochet%22:{%22version%22:6}}&shell_version=3.18.3
        return self.remote_meta['version'] > self.meta['version']

    def enabled(self):
        return self.uuid in (Gio.Settings
                             .new('org.gnome.shell')
                             .get_value('enabled-extensions'))

    def installed(self):
        installed = True
        try:
            self.meta['uuid']
        except FileNotFoundError:
            installed = False
        return installed

    def supported(self):
        try:
            for f in ['description', 'name', 'version']:
                assert f in self.meta, 'Missing field  - "{}"'.format(f)
            return True
        except Exception as e:
            warnings.warn("[{}] {}: metadata {}".format(self.uuid, e, self.meta))
            return False


class ExtensionManager:

    def __init__(self, ext_dir=EXTENSION_DIR):
        self.ext_dir = ext_dir

    def enabled_uuids(self):
        return set(
            Gio.Settings.new('org.gnome.shell')
            .get_value('enabled-extensions')
        ).intersection({ex.uuid for ex in self.installed()})

    def installed_uuids(self):
        return os.listdir(self.ext_dir)

    def installed(self):
        installed = []
        for uuid in self.installed_uuids():
            ext = Extension(uuid)
            if ext.supported():
                installed.append(ext)
        return installed

    def enabled(self):
        return [Extension(uuid) for uuid in self.enabled_uuids()]

    def disabled(self):
        installed_uuids = {e.uuid for e in self.installed()}
        enabled_uuids = {e.uuid for e in self.enabled()}
        return [Extension(uuid) for uuid in installed_uuids.difference(enabled_uuids)]

    def outdated(self):
        # TODO: use https://extensions.gnome.org/update-info/?installed={%22arch-update@RaphaelRochet%22:{%22version%22:6}}&shell_version=3.18.3
        return [e for e in self.enabled() if e.outdated()]

    def search(self, term):
        query = {
            'shell_version': '.'.join(str(v) for v in GNOME_SHELL_VERSION),
            'search': term,
        }
        response = get_json_response(API_SEARCH, query)
        found = []
        for remote_meta in response['extensions']:
            ext = Extension(remote_meta['uuid'])
            ext.remote_meta = remote_meta
            found.append(ext)
        return found

    def enable(self, uuid):
        if uuid in self.enabled_uuids():
            return True
        ext = Extension(uuid)
        if not ext.installed():
            return False
        enabled_uuids = self.enabled_uuids()
        enabled_uuids.add(uuid)
        Gio.Settings('org.gnome.shell').set_value(
            'enabled-extensions',
            GLib.Variant('as', list(enabled_uuids))
        )
        return True

    def disable(self, uuid):
        if uuid not in self.installed_uuids():
            raise Extension('Not installed')
        enabled_uuids = self.enabled_uuids()
        if uuid not in enabled_uuids:
            return True
        enabled_uuids.remove(uuid)
        Gio.Settings('org.gnome.shell').set_value(
            'enabled-extensions',
            GLib.Variant('as', list(enabled_uuids))
        )
        return True

    def install(self, uuid):
        ext = Extension(uuid)
        download_url = API_ROOT + ext.remote_meta['download_url']
        dest_dir = os.path.join(EXTENSION_DIR, uuid)
        download_and_extract_zip(download_url, dest_dir)
        return ext

    def uninstall(self, uuid):
        self.disable(uuid)
        shutil.rmtree(os.path.join(EXTENSION_DIR, uuid))



import urllib.parse
import urllib.request
import json
import subprocess

from io import BytesIO
from zipfile import ZipFile


def gnome_shell_version():
    """Get Gnome Shell version number.

    :returns: tuple(major, minor, patch)

    """
    cmd = subprocess.Popen(
            ['gnome-shell', '--version'],
            stdout=subprocess.PIPE
    )
    version_str = cmd.communicate()[0].split()[-1]
    return tuple(int(i) for i in version_str.split(b'.'))


GNOME_SHELL_VERSION = gnome_shell_version()


def get_json_response(endpoint, query):
    """Returns json response.

    :endpoint: url string
    :query: query dict
    :returns: response dict

    """
    query_string = urllib.parse.urlencode(query)
    full_url = "{}/?{}".format(endpoint, query_string)
    with urllib.request.urlopen(full_url) as f:
        data = f.read().decode('utf-8')
    return json.loads(data)


def download_and_extract_zip(url, dest):
    """Download zipfile and extract to destination dir.

    :url: url to download zipfile
    :dest: destination directory

    """
    with urllib.request.urlopen(url) as f:
        data = f.read()
    zipfile = ZipFile(BytesIO(data))
    zipfile.extractall(path=dest)


def reload_gnome_shell():
    """Reload gnome shell.

    """
    cmd = subprocess.Popen(
        "dbus-send --type=method_call --dest=org.gnome.Shell /org/gnome/Shell "
        "org.gnome.Shell.Eval string:'global.reexec_self()'",
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    cmd.communicate()



import argparse

from gsem.config import API_ROOT
from gsem.config import EXTENSION_DIR
from gsem.extension import Extension
from gsem.extension import ExtensionManager
from gsem.utils import reload_gnome_shell


def cli_args():
    """Parse command line arguments.

    :returns: argparse.Namespace with parsed args

    """
    parser = argparse.ArgumentParser(description='Gnome-Shell extension manager')
    commands = parser.add_subparsers(dest='cmd')
    commands.add_parser('ls', help='list installed extensions')
    commands.add_parser('enabled', help='list enabled extensions')
    commands.add_parser('disabled', help='list disabled extensions')
    commands.add_parser('outdated', help='list outdated extensions')
    info = commands.add_parser('info', help='show extension information')
    info.add_argument('uuid', help='extension uuid', metavar='UUID')
    info.add_argument('-r', '--remote', help='fetch info from {}'.format(API_ROOT),
                      action='store_true', default=False)
    install = commands.add_parser('install', help='install extension')
    install.add_argument('uuid', help='extension uuid')
    install.add_argument('--disabled', help='do not enable the extension', default=False,
                         action='store_true', dest='disabled')
    install.add_argument('--no-reload', help='do not reload gnome-shell', default=False,
                         action='store_true', dest='no_reload')
    reinstall = commands.add_parser('reinstall', help='reinstall extension')
    reinstall.add_argument('uuid', help='extension uuid', metavar='UUID')
    uninstall = commands.add_parser('uninstall', help='uninstall extension')
    uninstall.add_argument('uuid', help='extension uuid', metavar='UUID')
    commands.add_parser('update', help='update extensions')
    search = commands.add_parser('search', help='search extensions')
    search.add_argument('term', help='search term', metavar='TERM')
    enable = commands.add_parser('enable', help='enable extension')
    enable.add_argument('uuid', help='extension uuid', metavar='UUID')
    disable = commands.add_parser('disable', help='disable extension')
    disable.add_argument('uuid', help='extensions uuid', metavar='UUID')
    return parser


def print_nice_list(l):
    """Nicely print list.

    :l: list to print

    """
    length = len(l)
    for index, item in enumerate(l):
        if index < length-1:
            print('├── {}'.format(item))
        else:
            print('└── {}'.format(item))


def print_info(ext):
    """Print extension info.

    :ext: gsem.extension.Extension object

    """
    print("{} - ({})".format(ext.remote_meta['name'], ext.uuid))
    print()
    if ext.installed():
        if ext.outdated():
            print('outdated: {} -> {}'.format(ext.meta['version'], ext.remote_meta['version']))
        else:
            print('up to date: {}'.format(ext.meta['version']))
    else:
        print('not installed: {}'.format(ext.remote_meta['version']))
    print()
    print(ext.remote_meta['description'])


def main():
    # TODO: This is ugly!
    """Main cli function."""
    parser = cli_args()
    args = parser.parse_args()
    if not args.cmd:
        parser.print_usage()
        parser.exit(1)
    manager = ExtensionManager()
    list_cmd_map = {
        'disabled': manager.disabled,
        'enabled': manager.enabled,
        'ls': manager.installed,
        'outdated': manager.outdated,
        'search': lambda: manager.search(args.term),
    }
    if args.cmd in list_cmd_map.keys():
        result = list_cmd_map[args.cmd]()
        if args.cmd == 'outdated':
            print("{} ({})".format(manager.ext_dir, len(result)))
            l = ["{} {} -> {}".format(e.meta['uuid'], e.meta['version'], e.remote_meta['version']) for e in result]
        if args.cmd == 'search':
            print("Search results for '{}' ({})".format(args.term, len(result)))
            l = ['{} - {}'.format(e.remote_meta['uuid'], e.remote_meta['name']) for e in result]
        else:
            l = ["{}@{}".format(e.uuid, e.meta['version']) for e in result]
            print("{} ({})".format(manager.ext_dir, len(l)))
        print_nice_list(l)
    elif args.cmd == 'info':
        print_info(Extension(args.uuid))
    elif args.cmd == 'enable':
        enabled = manager.enable(args.uuid)
        if enabled:
            print("'{}' enabled".format(args.uuid))
        else:
            print("Can't enable '{}'".format(args.uuid))
    elif args.cmd == 'disable':
        disabled = manager.disable(args.uuid)
        if disabled:
            print("'{}' disabled".format(args.uuid))
        else:
            print("Can't disable '{}".format(args.uuid))
    elif args.cmd == 'install':
        if args.uuid in manager.installed_uuids():
            print("'{}' already installed".format(args.uuid))
        else:
            print("Installing {} to '{}'".format(args.uuid, EXTENSION_DIR))
            manager.install(args.uuid)
            if not args.disabled:
                print("Enabling '{}'".format(args.uuid))
                manager.enable(args.uuid)
                if not args.no_reload:
                    print("Reloading gnome-shell...")
                    reload_gnome_shell()
    elif args.cmd == 'reinstall':
        if args.uuid not in manager.installed_uuids():
            print("'{}' is not installed".format(args.uuid))
        else:
            print("Uninstalling '{}'".format(args.uuid))
            ext = Extension(args.uuid)
            was_enabled = ext.enabled()
            manager.uninstall(args.uuid)
            print("Installing {} to '{}'".format(args.uuid, EXTENSION_DIR))
            manager.install(args.uuid)
            if was_enabled:
                manager.enable(args.uuid)
    elif args.cmd == 'uninstall':
        if args.uuid not in manager.installed_uuids():
            print("'{}' is not installed".format(args.uuid))
        else:
            print("Uninstalling '{}'".format(args.uuid))
            manager.uninstall(args.uuid)
    elif args.cmd == 'update':
        outdated = manager.outdated()
        print('Extension updates avaliable ({})'.format(len(outdated)))
        if outdated:
            l = ["{} {} -> {}".format(e.uuid, e.meta['version'], e.remote_meta['version']) for e in outdated]
            print_nice_list(l)
            prompt = input('Would you like to install these updates? (yes) ')
            if prompt.lower().strip().startswith('y'):
                for e in outdated:
                    was_enabled = e.enabled()
                    print("Installing {}@{} to {}".format(e.uuid, e.remote_meta['version'], EXTENSION_DIR))
                    manager.uninstall(e.uuid)
                    manager.install(e.uuid)
                    if was_enabled:
                        manager.enable(e.uuid)
                reload_gnome_shell()



import os
from gsem.utils import gnome_shell_version

EXTENSION_DIR = os.path.expanduser(
    '~/.local/share/gnome-shell/extensions'
)
GNOME_SHELL_VERSION = gnome_shell_version()

API_ROOT = 'https://extensions.gnome.org'
API_DETAIL = API_ROOT + '/ajax/detail'
API_SEARCH = API_ROOT + '/extension-query'



from setuptools import setup


setup(name='gsem',
      version='0.1.1',
      description='Command line extension manager for Gnome-Shell',
      url='https://github.com/andriykohut/gsem',
      author='Andriy Kogut',
      author_email='kogut.andriy@gmail.com',
      license='MIT',
      packages=['gsem'],
      keywords='gsem gnome gnome-shell extention manager',
      zip_safe=True,
      entry_points={
          'console_scripts': ['gsem=gsem.cli:main'],
      })
